@@ -9,7 +9,7 @@ use MIDI::RtMidi::FFI ':all';
 use MIDI::Event;
 use Carp;
 
-our $VERSION = '0.04';
+our $VERSION = '0.049';
 
 my $rtmidi_api_names = {
     unspecified => [ "Unknown",            RTMIDI_API_UNSPECIFIED ],
@@ -140,9 +140,31 @@ sub cancel_callback {
 
 
 sub ignore_types {
-    my ( $self, $sysex, $time, $sense ) = @_;
+    my ( $self, $sysex, $timing, $sensing ) = @_;
+    @{ $self }{ qw/ ignore_sysex ignore_timing ignore_sensing / } = ( $sysex, $timing, $sensing );
     croak "Unable to ignore_types for device type : $self->{type}" unless $self->{type} eq 'in';
-    rtmidi_in_ignore_types( $self->{device}, $sysex, $time, $sense );
+    rtmidi_in_ignore_types( $self->{device}, $sysex, $timing, $sensing );
+}
+
+
+sub ignore_sysex {
+    my ( $self, $ignore_sysex ) = @_;
+    $self->{ignore_sysex} = $ignore_sysex;
+    $self->ignore_types( @{ $self }{ qw/ ignore_sysex ignore_timing ignore_sensing / } )
+}
+
+
+sub ignore_timing {
+    my ( $self, $ignore_timing ) = @_;
+    $self->{ignore_timing} = $ignore_timing;
+    $self->ignore_types( @{ $self }{ qw/ ignore_sysex ignore_timing ignore_sensing / } )
+}
+
+
+sub ignore_sensing {
+    my ( $self, $ignore_sensing ) = @_;
+    $self->{ignore_sensing} = $ignore_sensing;
+    $self->ignore_types( @{ $self }{ qw/ ignore_sysex ignore_timing ignore_sensing / } )
 }
 
 
@@ -268,6 +290,24 @@ sub DESTROY {
     # $fn->( $self->{device} );
 }
 
+{
+    package RtMidiIn;
+    use strict; use warnings;
+    sub new {
+        shift;
+        MIDI::RtMidi::FFI::Device->new( @_, type => 'in' );
+    }
+}
+
+{
+    package RtMidiOut;
+    use strict; use warnings;
+    sub new {
+        shift;
+        MIDI::RtMidi::FFI::Device->new( @_, type => 'out' );
+    }
+}
+
 1;
 
 __END__
@@ -282,13 +322,13 @@ MIDI::RtMidi::FFI::Device - OO interface for MIDI::RtMidi::FFI
 
 =head1 VERSION
 
-version 0.04
+version 0.049
 
 =head1 SYNOPSIS
 
     use MIDI::RtMidi::FFI::Device;
     
-    my $device = MIDI::RtMidi::FFI::Device->new;
+    my $device = RtMidiOut->new;
     $device->open_virtual_port( 'perl-rtmidi' );
     $device->send_event( note_on => 0x00, 0x40, 0x5a );
     sleep 1;
@@ -303,9 +343,13 @@ you manage devices, ports and MIDI events.
 
 =head2 new
 
-    my $device = MIDI::RtMidi::FFI::Device->new( %attributes );
+    my $device = MIDI::RtMidi::FFI::Device->new( %options );
+    my $midiin = RtMidiIn->new( %options );
+    my $midiout = RtMidiOut->new( %options );
 
-Returns a new MIDI::RtMidi::FFI::Device object. Valid attributes:
+Returns a new MIDI::RtMidi::FFI::Device object. RtMidiIn and RtMidiOut are
+provided as shorthand to instantiate devices of type 'in' and 'out'
+respectively. Valid attributes:
 
 =over 4
 
@@ -314,6 +358,8 @@ Returns a new MIDI::RtMidi::FFI::Device object. Valid attributes:
 B<type> -
 Device type : 'in' or 'out' (defaults to 'out')
 
+Ignored if instantiating RtMidiIn or RtMidiOut.
+
 =item *
 
 B<api> -
@@ -478,6 +524,27 @@ Type 'in' only. Removes the callback from your device.
 
 Type 'in' only. Set message types to ignore.
 
+=head2 ignore_sysex
+
+    $device->ignore_sysex( 1 );
+    $device->ignore_sysex( 0 );
+
+Type 'in' only. Set whether or not to ignore sysex messages.
+
+=head2 ignore_timing
+
+    $device->ignore_timing( 1 );
+    $device->ignore_timing( 0 );
+
+Type 'in' only. Set whether or not to ignore timing messages.
+
+=head2 ignore_sensing
+
+    $device->ignore_sensing( 1 );
+    $device->ignore_sensing( 0 );
+
+Type 'in' only. Set whether or not to ignore active sensing messages.
+
 =head2 get_message
 
     $device->get_message();
@@ -570,7 +637,7 @@ John Barrett <john@jbrt.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2021 by John Barrett.
+This software is copyright (c) 2023 by John Barrett.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
